/* This file is part of the KDE project
   Copyright (C) 2002 Simon Hausmann <hausmann@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "frontend.h"
#include "messagedispatcher.h"
#include "connection.h"
#include "window.h"
#include "windowforwarder.h"
#include "messagefactory.h"
#include "commands.h"
#include "messagefilter.h"
#include "standardmessagefilters.h"
#include "markup.h"
#include "ctcp.h"

#include <qvariant.h>
#include <qtimer.h>

#include <klocale.h>
#include <klineeditdlg.h>

#include <assert.h>

using namespace IRC;

Frontend::Frontend( Connection *connection, const ConnectionInfo &connInfo,
                    QObject *parent, const char *name )
    : QObject( parent, name ), m_connection( connection ), m_connInfo( connInfo ), m_userUpdateRunning( false )
{
    m_dispatcher = new MessageDispatcher( this );
    connect( m_dispatcher, SIGNAL( unhandledMessage( const IRC::Message & ) ),
             this, SLOT( unhandledMessage( const IRC::Message & ) ) );

    m_dispatcher->registerMessageReceiver( this );

    m_forwarder = new WindowForwarder( this );
    connect( m_forwarder, SIGNAL( unhandledMessage( const IRC::Message & ) ),
             this, SLOT( unhandledMessage( const IRC::Message & ) ) );
    m_dispatcher->registerMessageReceiver( m_forwarder );

    connect( m_connection, SIGNAL( connected() ),
             this, SLOT( connected() ) );
    connect( m_connection, SIGNAL( connectionClosed() ),
             this, SLOT( connectionClosed() ) );
    connect( m_connection, SIGNAL( lineRead( const QString & ) ),
             this, SLOT( processInput( const QString & ) ) );

    registerInternalMessageTypes();

    m_messageFilter = new MessageFilterGroup( this );
    m_namedMessageFilters = new NamedMessageFilterGroup( m_messageFilter );

    registerStandardMessageFilters();

    m_ctcpHandler = new CTCPHandler( this );
}

Frontend::~Frontend()
{
    m_windows.setAutoDelete( true );
    m_windows.clear();
}

Window *Frontend::windowForChannel( const QString &channel, WindowCreationPolicy winPolicy )
{
    Window *wnd = m_windows.find( channel );

    if ( wnd )
        return wnd;

    if ( winPolicy == CreateIfNotExisting ) {
        wnd = createWindow( channel );

        m_windows.insert( channel, wnd );

        m_forwarder->registerWindow( wnd );

        connectWindow( wnd );
    }
    else if ( winPolicy == ReturnDefaultWindow )
        wnd = defaultWindow();

    return wnd;
}

Window *Frontend::defaultWindow()
{
    if ( !m_defaultWindow ) {
        m_defaultWindow = createDefaultWindow();
        connectWindow( m_defaultWindow );
    }
    return m_defaultWindow;
}

bool Frontend::hasWindow( const QString &channel ) const
{
    return m_windows.find( channel ) != 0;
}

Message Frontend::createPrefixedMessage( const QString &type, const QStringList &parameters )
{
    QString prefix = m_connInfo.toMaskString();
    return Message( prefix, type, parameters );
}

Person Frontend::userJoinedChannel( const QString &nick, Window *window )
{
    PersonList::Iterator personIt = m_users.find( Person( nick ) );
    Person person;
    if ( personIt == m_users.end() ) {
        person.setNick( nick );
        personIt = m_users.append( person );
    } else
        person = *personIt;

    person.channels()[ window->channel() ] = ChannelUserAttributes();

    return person;
}

void Frontend::userLeftChannel( const QString &nick, Window *window )
{
    PersonList::Iterator personIt = m_users.find( Person( nick ) );
    if ( personIt == m_users.end() )
        return;

    Person &person = *personIt;
    person.channels().remove( window->channel() );

    Person::ChannelInfoMap channels = person.channels();
    for ( Person::ChannelInfoMap::ConstIterator channelIt = channels.begin();
          channelIt != channels.end(); ++channelIt ) 
        if ( hasWindow( channelIt.key() ) ) 
            return;

    m_users.remove( Person( nick ) );

    m_pendingUserUpdates.remove( nick );
}

void Frontend::installReplyHandler( const QCString &messageType, QObject *handler, MessageDispatcher::MessageReceiverType receiverType )
{
    connectReplyHandler( messageType, handler, receiverType );
}

void Frontend::sendOneWayMessage( const IRC::Message &msg )
{
    if ( !m_connection )
        defaultWindow()->printMessage( i18n( "Not connected to a server" ) );

    // ### hack. don't confuse IRC servers by sending prefixed messages.
    Message rawMsg( QString::null, msg.type(), msg.parameters() );

    m_connection->writeLine( rawMsg.serialize() );
}

void Frontend::sendMessage( const IRC::Message &msg, QObject *replyHandler )
{
    QCString type = msg.type().latin1();
    connectReplyHandler( type, replyHandler );

    sendOneWayMessage( msg );
}

void Frontend::executeCommand( IRC::Command *cmd )
{
    connectCommand( cmd );
    cmd->start();
}

void Frontend::unhandledMessage( const IRC::Message &msg )
{
    if ( msg.isNumericType() && !msg.parameters().isEmpty() ) {
        // we assume it's an unknown numeric reply in the form of
        // :<prefix> <type> <yournickname> <params>
        // Let's just display the prefix and the params, they usually give good information
        QStringList params = msg.parameters();
        params.pop_front(); 
        // ### Global::formatErrorMessage( msgBody );
        QString txt = "[" + msg.prefix() + "] " + params.join( " " );
        defaultWindow()->printMessage( Markup::tagError( txt ) );
        return;
    }

    // ###
    if ( msg.type() == NoticeMessage::name() || PrivmsgMessage::name() ) {
        PrivmsgMessage pmsg( msg );

        QString txt = "[" + msg.prefix() + "] " + pmsg.plainText();

        defaultWindow()->printMessage( txt );
        return;
    }

    defaultWindow()->printMessage( i18n( "Received unknown command: %1" ).arg( msg.serialize() ) );
}

void Frontend::updateUserDetails( const PersonList &users )
{
    for ( PersonList::ConstIterator it = users.begin(); it != users.end(); ++it )
        updateUserDetails( ( *it ).nick() );
}

void Frontend::updateUserDetails( const QString &nick )
{
    if ( m_pendingUserUpdates.findIndex( nick ) == -1 ) 
        m_pendingUserUpdates << nick;

    if ( !m_userUpdateRunning ) 
        nextUserUpdateDelayed();
}

void Frontend::windowDestroyed( Window *window )
{
    assert( hasWindow( window->channel() ) );
    m_windows.remove( window->channel() );
}

void Frontend::processInput( const QString &input )
{
    Message msg = Message::fromString( input );
    if ( !msg.isValid() ) {
        defaultWindow()->printMessage( i18n( "Received invalid message from server: %1" ).arg( input ) );
        return;
    }

    m_dispatcher->dispatchMessage( msg );
}

void Frontend::connected()
{
    defaultWindow()->printMessage( i18n( "Connected to %1" ).arg( m_connection->peerDescription() ) );

    Command *cmd = new LoginCommand( m_connection->address(), defaultWindow() );
    connect( cmd, SIGNAL( commandFinished( IRC::Command * ) ),
             this, SIGNAL( loggedIn() ) );
    executeCommand( cmd );
}

void Frontend::connectionClosed()
{
    defaultWindow()->printMessage( i18n( "Connection to server lost!" ) );
}

void Frontend::handlePingMessage( const IRC::PingMessage &ping )
{
    sendOneWayMessage( ping.pong() );
}

void Frontend::handleQuitMessage( const IRC::QuitMessage &msg )
{
    PersonList::Iterator personIt = m_users.find( msg.sender() );
    if ( personIt == m_users.end() )
        return;

    Person person = *personIt;

    Person::ChannelInfoMap channels = person.channels();
    for ( Person::ChannelInfoMap::ConstIterator it = channels.begin();
          it != channels.end(); ++it ) {

        Window *window = windowForChannel( it.key(), ReturnNullIfNotExisting );
        if ( !window )
            continue;

        window->handleQuitMessage( msg );
    }

    m_users.remove( personIt );
}

void Frontend::handleNickMessage( const IRC::NickMessage &msg )
{
    PersonList::Iterator person = m_users.find( msg.sender() );
    if ( person == m_users.end() )
        return;

    ( *person ).setNick( msg.nick() );

    // ### update nick in connectioninfo

    Person::ChannelInfoMap channels = ( *person ).channels();
    for ( Person::ChannelInfoMap::ConstIterator it = channels.begin();
          it != channels.end(); ++it ) {

        Window *window = windowForChannel( it.key(), ReturnNullIfNotExisting );
        if ( !window )
            continue;

        window->handleNickMessage( msg );
    }
}

void Frontend::unregisterCommand( IRC::Command *cmd )
{
    m_dispatcher->unregisterMessageReceiver( cmd );
}

void Frontend::nextUserUpdate()
{
    QTimer::singleShot( USERDETAILS_UPDATE_DELAY, this, SLOT( nextUserUpdateDelayed() ) );
}

void Frontend::nextUserUpdateDelayed()
{
    if ( m_pendingUserUpdates.isEmpty() ) {
        m_userUpdateRunning = false;
        return;
    }

    m_userUpdateRunning = true;

    ReplyHandler *handler = new WhoisReplyHandler( defaultWindow() );
    connect( handler, SIGNAL( userDetailsAvailable( const IRC::Person & ) ),
             this, SLOT( updateUserDetails( const IRC::Person & ) ) );

    QString nick = m_pendingUserUpdates.first();
    m_pendingUserUpdates.remove( m_pendingUserUpdates.first() );

    WhoisMessage msg;
    msg.setTarget( nick );

    m_userUpdateCommand = new SimpleCommand( msg, handler, defaultWindow() );
    connect( m_userUpdateCommand, SIGNAL( commandFinished( IRC::Command * ) ),
             this, SLOT( nextUserUpdate() ) );

    executeCommand( m_userUpdateCommand );
}

void Frontend::updateUserDetails( const IRC::Person &userDetails )
{
    PersonList::Iterator userIt = m_users.find( userDetails );
    if ( userIt == m_users.end() )
        return;

    Person user = *userIt;
    user.copyFrom( userDetails );

    emit userDetailsAvailable( user );
}

void Frontend::connectCommand( Command *command )
{
    connect( command, SIGNAL( sendMessage( const IRC::Message &, QObject * ) ),
             this, SLOT( sendMessage( const IRC::Message &, QObject * ) ) );

    connect( command, SIGNAL( commandFinished( IRC::Command * ) ),
             this, SLOT( unregisterCommand( IRC::Command * ) ) );
}

void Frontend::connectReplyHandler( const QCString &msgType, QObject *replyHandler, MessageDispatcher::MessageReceiverType receiverType )
{
    Message::NumericReplyInfo replyInfo = MessageFactory::self().replyInfo( msgType );

    QVariant usesTerminatingCommandSequence = replyHandler->property( "usesTerminatingCommandSequence" );
    if ( usesTerminatingCommandSequence.isValid() && usesTerminatingCommandSequence.toBool() )
        receiverType = MessageDispatcher::PermanentMessageReceiver;

    connectReplyHandler( replyHandler, replyInfo.errorCodes, receiverType );
    connectReplyHandler( replyHandler, replyInfo.replyCodes, receiverType );

    m_dispatcher->registerMessageReceiver( replyHandler, receiverType );
}

void Frontend::connectReplyHandler( QObject *replyHandler, const Message::NumericMessageTypeArray &replyCodes, MessageDispatcher::MessageReceiverType receiverType )
{
    Message::NumericMessageTypeArray::ConstIterator it = replyCodes.begin();
    for ( ; it != replyCodes.end(); ++it ) {
        QString stringifiedCode = numericReplyToString( *it );

        m_dispatcher->registerMessageReceiver( stringifiedCode, replyHandler, receiverType );
    }
}

void Frontend::registerInternalMessageTypes()
{
    Message::NumericMessageTypeArray codes( 2 );
    codes[ 0 ] = RPL_TOPIC;
    codes[ 1 ] = RPL_NOTOPIC;
    connectReplyHandler( m_forwarder, codes, MessageDispatcher::PermanentMessageReceiver );
}

void Frontend::registerStandardMessageFilters()
{
    m_messageFilter->addFilter( m_namedMessageFilters );

    m_namedMessageFilters->addFilter( new ActionFilter( this ) );
    m_namedMessageFilters->addFilter( new PartFilter( this ) );
    m_namedMessageFilters->addFilter( new QueryFilter( this ) );
    m_namedMessageFilters->addFilter( new QuoteFilter( this ) );
    m_namedMessageFilters->addFilter( new TopicFilter( this ) );
}

void Frontend::connectWindow( Window *wnd )
{
    connect( wnd, SIGNAL( windowDestroyed( Window * ) ),
             this, SLOT( windowDestroyed( Window * ) ) );
    connect( wnd, SIGNAL( textEntered( Window *, const QString &input ) ),
             this, SLOT( processInput( Window *, const QString &input ) ) );
    connect( wnd, SIGNAL( sendMessage( const IRC::Message &, QObject * ) ),
             this, SLOT( sendMessage( const IRC::Message &, QObject * ) ) );
    connect( wnd, SIGNAL( sendOneWayMessage( const IRC::Message & ) ),
             this, SLOT( sendOneWayMessage( const IRC::Message & ) ) );
    connect( wnd, SIGNAL( executeCommand( IRC::Command * ) ),
             this, SLOT( executeCommand( IRC::Command * ) ) );
}

#include "frontend.moc"

/* vim: et sw=4 ts=4
 */
