/* This file is part of the KDE project
   Copyright (C) 2002 Simon Hausmann <hausmann@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#ifndef GLOBAL_H
#define GLOBAL_H

#include <qstring.h>
#include <qstringlist.h>
#include <qmap.h>

namespace IRC
{

    void initStatic();

enum
{
    // time after join when first NAMES update is sent
    INITIAL_USERLIST_UPDATE_TIMEOUT = 1 * 1000,
    // update userlist every minute
    USERLIST_UPDATE_INTERVAL = 60 * 1000,
    // pause between automated WHOIS sequences sent for user details
    USERDETAILS_UPDATE_DELAY = 5 * 1000
};

struct ServerInfo
{
    ServerInfo() {}
    ServerInfo( const QString &_hostName, const QString &_description = QString::null )
        : hostName( _hostName ), description( _description )
    {}

    QString hostName;
    QString description;
};

inline bool operator==( const ServerInfo &lhs, const ServerInfo &rhs )
{ return lhs.hostName == rhs.hostName; }

inline bool operator<( const ServerInfo &lhs, const ServerInfo &rhs )
{ return lhs.hostName < rhs.hostName; }

enum ChannelType { SecretChannel, PrivateChannel, PublicChannel };

struct ChannelUserAttributes
{
    ChannelUserAttributes() : isChannelOperator( false ), hasVoice( false ) {}
    bool isChannelOperator : 1;
    bool hasVoice : 1;
};

struct PersonData;

class Person
{
public:
    typedef QMap<QString, ChannelUserAttributes> ChannelInfoMap;

    Person();
    explicit Person( const QString &nick );
    Person( const QString &nick, const QString &user, const QString &realName, const QString &host );
    Person( const Person &rhs );
    Person &operator=( const Person &rhs );
    ~Person();

    QString nick() const;
    QString user() const;
    QString realName() const;
    QString host() const;
    ServerInfo ircServer() const;
    bool isIRCOperator() const;
    bool isAway() const;
    QString awayMessage() const;
    int idleTime() const;
    const ChannelInfoMap &channels() const;
    ChannelInfoMap &channels();

    void setNick( const QString &nick );
    void setUser( const QString &user );
    void setRealName( const QString &realName );
    void setHost( const QString &host );
    void setIRCServer( const ServerInfo &server );
    void setIRCOperator( bool _operator );
    void setAwayMessage( const QString &awayMessage );
    void setIdleTime( int idleTimeInSeconds );

    bool isNull() const;

    bool operator==( const Person &rhs ) const;
    bool operator<( const Person &rhs ) const;

    /* a mask string is like nick!user@host */
    QString toMaskString() const;
    static Person fromMaskString( const QString &mask );

    /* return format: nick!user@host (realname) */
    QString toWhoisString() const;

    void copyFrom( const Person &other );

private:
    PersonData *d;
};

struct PersonData : public QShared
{
    PersonData() : m_ircOp( false ), m_idleTime( -1 ) {}
    PersonData( const QString &nick, const QString &user, const QString &realName, const QString &host )
        : m_nick( nick ), m_user( user ), m_realName( realName ), m_host( host ), m_ircOp( false ), m_idleTime( -1 )
    {}

    void copyFrom( const PersonData &other );

    QString m_nick;
    QString m_user;
    QString m_realName;
    QString m_host;
    ServerInfo m_ircServer;
    bool m_ircOp;
    QString m_awayMessage;
    int m_idleTime;
    Person::ChannelInfoMap m_channels;
};

inline Person::Person() : d( new PersonData ) {}

inline Person::Person( const QString &nick ) : d( new PersonData ) 
{
    d->m_nick = nick;
}

inline Person::Person( const QString &nick, const QString &user, const QString &realName, const QString &host )
    : d( new PersonData( nick, user, realName, host ) )
{}

inline Person::Person( const Person &rhs )
    : d( new PersonData )
{ (*this) = rhs; }

inline Person &Person::operator=( const Person &rhs )
{
    rhs.d->ref();

    if ( d->deref() ) delete d;

    d = rhs.d;

    return *this;
}

inline Person::~Person()
{
    if ( d->deref() )
        delete d;
}

inline QString Person::nick() const { return d->m_nick; }
inline QString Person::user() const { return d->m_user; }
inline QString Person::realName() const { return d->m_realName; }
inline QString Person::host() const { return d->m_host; }
inline ServerInfo Person::ircServer() const { return d->m_ircServer; }
inline bool Person::isIRCOperator() const { return d->m_ircOp; }
inline bool Person::isAway() const { return !d->m_awayMessage.isNull(); }
inline QString Person::awayMessage() const { return d->m_awayMessage; }
inline int Person::idleTime() const { return d->m_idleTime; }
inline const Person::ChannelInfoMap &Person::channels() const { return d->m_channels; }
inline Person::ChannelInfoMap &Person::channels() { return d->m_channels; }

inline void Person::setNick( const QString &nick ) { d->m_nick = nick; }
inline void Person::setUser( const QString &user ) { d->m_user = user; }
inline void Person::setRealName( const QString &realName ) { d->m_realName = realName; }
inline void Person::setHost( const QString &host ) { d->m_host = host; }
inline void Person::setIRCServer( const ServerInfo &server ) { d->m_ircServer = server; }
inline void Person::setIRCOperator( bool _operator ) { d->m_ircOp = _operator; }
inline void Person::setAwayMessage( const QString &awayMessage ) { d->m_awayMessage = awayMessage; }
inline void Person::setIdleTime( int idleTimeInSeconds ) { d->m_idleTime = idleTimeInSeconds; }

inline bool Person::isNull() const 
{ return d->m_nick.isNull() && d->m_user.isNull() && d->m_realName.isNull() && d->m_host.isNull(); }

inline bool Person::operator==( const Person &rhs ) const
{ return d->m_nick == rhs.d->m_nick; }

inline bool Person::operator<( const Person &rhs ) const
{ return d->m_nick < rhs.d->m_nick; }

inline QString Person::toMaskString() const
{ return d->m_nick + '!' + d->m_user + '@' + d->m_host; }

inline QString Person::toWhoisString() const
{ return toMaskString() + " (" + d->m_realName + ")"; }

inline void Person::copyFrom( const Person &other )
{ d->copyFrom( *other.d ); }

typedef QValueList<Person> PersonList;
typedef QMap<QString, Person> PersonMap;

inline QString numericReplyToString( int code )
{
    return QString::number( code ).rightJustify( 3, '0' );
}

QStringList splitStringListObeyingQuotes( QChar separator, const QString &text );

} // namespace IRC

#endif // GLOBAL_H
/* vim: et sw=4 ts=4
 */
